/*------------------------------------------------------------------------------
 * Copyright (C) 1993 Christian-Albrechts-Universitaet zu Kiel
 *------------------------------------------------------------------------------
 * Projekt  : APPLY - A Practicable And Portable Lisp Implementation
 *            ------------------------------------------------------
 * Funktion : Laufzeitsystem-Routinen: Speicherverwaltung, Abort-Funktion
 *
 * $Revision: 1.16 $
 * $Log: system.c,v $
 * Revision 1.16  1993/07/28  15:05:28  hk
 * Bug in char_alloc behoben: das abschliessende 0 Character wurde nicht
 * direkt hinter dem String sondern um 1 zu weit dahinter geschrieben.
 *
 * Revision 1.15  1993/07/13  15:26:31  sma
 * Kopieren durch Systemaufruf memcpy
 *
 * Revision 1.14  1993/06/16  14:43:22  hk
 * Copyright Notiz eingefuegt.
 *
 * Revision 1.13  1993/05/31  17:08:31  pm
 * Fehler in den c-floats entfernt
 *
 * Revision 1.12  1993/05/23  17:55:04  pm
 * die Funktion safe_form um den Test fuer die C-Typen erweitert
 *
 * Revision 1.11  1993/05/03  12:35:39  pm
 * garbage-collector versteht jetzt auch die Tags des FFI
 *
 * Revision 1.10  1993/04/22  10:27:26  hk
 * Neue Funktion gc_symbols, die die Symbole eines Moduls traversiert.
 * Bei der Garbage Collection wird gc_main aufgerufen, um das Traversieren
 * der Symbole der Module zu veranlassen. gc_main wird im Hauptprogramm
 * definiert.
 *
 * Revision 1.9  1993/03/25  09:06:38  hk
 * Fehlermeldungen und GC Meldungen nach stderr.
 *
 * Revision 1.8  1993/02/26  14:54:39  ft
 * Den GC um die Behandlung von Instanzen erweitert.
 *
 * Revision 1.7  1993/02/17  15:49:41  hk
 * CLICC -> APPLY, Revison Keyword.
 *
 * Revision 1.6  1993/01/06  15:45:26  hk
 * *** empty log message ***
 *
 * Revision 1.5  1993/01/06  15:41:48  hk
 * Character Arrays sind mit einem 0 Character abgeschlossen.
 *
 * Revision 1.4  1992/10/05  16:09:34  uho
 * Statt calloc jetzt malloc benutzt, da calloc mit 0 initialisiert, was
 * gar nicht notwendig ist.
 * Die MIN* und MAX* Preprocessor Symbole in externe Variablen umgewandelt.
 * Sie werden nun im erzeugten Programm definiert.
 *
 * Revision 1.3  1992/10/02  14:58:22  uho
 * Definition der Heapbereiche auf dynamisches Anfordern umgestellt.
 * memallot zum dynamischen Anfordern des Heaps angelegt.
 * Groessen der einzelnen Heapbereiche statt als Konstanten nun als globale
 * Variablen definiert.
 *
 * Revision 1.2  1992/09/28  17:20:28  hk
 * Lerror -> Labort, neues Lerror mit Lisp-Parameter
 *
 * Revision 1.1  1992/03/24  17:03:37  hk
 * Initial revision
 *----------------------------------------------------------------------------*/
#ifdef __STDC__
#include <stdlib.h>
#else
#include <malloc.h>
#endif
#include "c_decl.h"
#include "sys.h"

/*------------------------------------------------------------------------------
 * The sizes of the four memory areas used by the runtime system  will be
 * defined in the generated program as initialized global variables.
 *----------------------------------------------------------------------------*/

extern unsigned maxstacksize;
extern unsigned minstacksize;

extern unsigned maxform_heapsize;
extern unsigned minform_heapsize;

extern unsigned maxfx_heapsize;
extern unsigned minfx_heapsize;

extern unsigned maxfl_heapsize;
extern unsigned minfl_heapsize;

/*------------------------------------------------------------------------------
 * LISP-Laufzeitstack
 *----------------------------------------------------------------------------*/
CL_FORM *stack;
CL_FORM *eos;              /* end of stack */
unsigned stacksize;

/*------------------------------------------------------------------------------
 * LISP-Heap fuer CL_FORMs
 *----------------------------------------------------------------------------*/
CL_FORM *fo_heap;
CL_FORM *form_heap;
CL_FORM *form_toh;
CL_FORM *form_eoh;
CL_FORM *old_form_heap;
unsigned form_heapsize;

/*------------------------------------------------------------------------------
 * LISP-Heap fuer INTEGERs und STRINGs
 *----------------------------------------------------------------------------*/
long *fx_heap1;

long *fx_heap;
long *fx_toh;
long *fx_eoh;

long *old_fx_heap;

unsigned fx_heapsize;


/*------------------------------------------------------------------------------
 * LISP-Heap fuer FLOATs
 *----------------------------------------------------------------------------*/
double *fl_heap1;

double *fl_heap;
double *fl_toh;
double *fl_eoh;

double *old_fl_heap;

unsigned fl_heapsize;

/*------------------------------------------------------------------------------
 * void *allocate(nelem, minelem, elsize, actelem)
 * unsigned nelem, minelem, elsize, *actnelem;
 *
 * Try to allocate a block of memory containing elements of size ELSIZE
 * Allocates between MINELEM and NELEM elements.
 * Returns a pointer to the newly allocated block, or NULL if not succesful
 * ACTELEM is assigne to the number of elements actually allocated.
 *----------------------------------------------------------------------------*/
static void *allocate(nelem, minelem, elsize, actnelem)
unsigned nelem, minelem, elsize, *actnelem;
{
   void *heap;

   while ((heap = malloc(nelem * elsize)) == NULL)
   {
      nelem /= 2;
      if (nelem < minelem)
      {  /* less than required, try minimal requirements */
         if ((heap = malloc(minelem * elsize)) == NULL)
         {
            *actnelem = 0;
            return NULL;
         }
         *actnelem = minelem;
         return heap;
      }
   }
   *actnelem = nelem;
   return heap;
}


/*------------------------------------------------------------------------------
 * Fehlermeldung
 *----------------------------------------------------------------------------*/
static void no_mem(msg, amount)
char *msg;
unsigned amount;
{
   fprintf(stderr, "Cannot allocate %u Bytes for %s!\n", amount, msg);
   Labort("Out of Memory!");
}


/*------------------------------------------------------------------------------
 * void memallot()
 *
 *  Alloctes memory for the different regions needed by the runtime
 *  system.
 *  Assigns the global veriables
 *
 *  Stack     : stack, stacksize, eos,
 *  FormHeap  : fo_heap, form_heapsize, form_toh, fo_heap, form_eoh,
 *  IntStrHeap: fx_heap1, fx_heapsize, fx_heap, fx_toh, fx_eoh, old_fx_heap,
 *  FloatHep  : fl_heap1, fl_heapsize, fl_heap, fl_toh, fl_eoh, old_fl_heap
 *
 *  If not enough memory can be allocated, MEMALLOT will raise an error and
 *  will abort program execution.
 *
 *----------------------------------------------------------------------------*/
void memallot()
{
   /* allocate stack */
   /*----------------*/
   stack = (CL_FORM *)allocate(maxstacksize, minstacksize,
                               sizeof(CL_FORM), &stacksize);
   if(stack == NULL)
      no_mem("Stack", minstacksize*sizeof(CL_FORM));
   eos = stack + stacksize;

   /* allocate form heap */
   /*--------------------*/
   fo_heap = (CL_FORM *)allocate(2 * maxform_heapsize, 2 * minform_heapsize,
                                 sizeof(CL_FORM), &form_heapsize);
   if(fo_heap == NULL) 
      no_mem("FormHeap", 2*minform_heapsize*sizeof(CL_FORM));

   form_heapsize /= 2;
   form_heap = form_toh = fo_heap;
   form_eoh = old_form_heap = fo_heap + form_heapsize;

   /* allocate integer- and string heap */
   /*-----------------------------------*/
   fx_heap1 = (long *)allocate(2 * maxfx_heapsize, 2 * minfx_heapsize,
                               sizeof(long), &fx_heapsize);
   if(fx_heap1 == NULL) 
      no_mem("Integer/StringHeap", 2*minfx_heapsize*sizeof(long));

   fx_heapsize /= 2;
   fx_heap = fx_toh = fx_heap1;
   fx_eoh = old_fx_heap = fx_heap1 + fx_heapsize;

   /* allocate float heap */
   /*---------------------*/
   fl_heap1 = (double *)allocate(2 * maxfl_heapsize, 2 * minfl_heapsize,
                                 sizeof(double), &fl_heapsize);
   if(fl_heap1 == NULL)
      no_mem("FloatHeap", 2*minfl_heapsize*sizeof(double));

   fl_heapsize /= 2;
   fl_heap = fl_toh = fl_heap1;
   fl_eoh = old_fl_heap = fl_heap1 + fl_heapsize;
}


/*----------------------------------------------------------------------------*/
#define GC_FORWARD    128
#define LOAD_FORWARD(fw_ptr, loc) \
   (SET_TAG (loc, GC_FORWARD), GET_FORM (loc) = (fw_ptr))


/*------------------------------------------------------------------------------
 * Kopiert 'num' FORMs in den neuen Heap.
 *----------------------------------------------------------------------------*/
CL_FORM *form_swap (from, num)
CL_FORM *from;
long num;
{
   CL_FORM *new_location = form_toh;

   memcpy(form_toh, from, num * sizeof(CL_FORM));
   form_toh += num;
   return new_location;
}

/*------------------------------------------------------------------------------
 * Kopiert 'num' FIXNUMs in den neuen Heap.
 *----------------------------------------------------------------------------*/
long *fx_swap (from, num)
long *from;
long num;
{
   long *new_location = fx_toh;

   memcpy(fx_toh, from, num * sizeof(long));
   fx_toh += num;
   return new_location;
}

/*------------------------------------------------------------------------------
 * Kopiert 'num' + 1 CHARACTERs in den neuen Heap.
 * (einschliesslich des abschliessenden 0 Characters)
 *----------------------------------------------------------------------------*/
char *ch_swap (from, num)
char *from;
long num;
{
   char *new_location = (char *)fx_toh;

   /* Auf nchstens Vielfaches von sizeof(long) runden */
   /* ANNAHME: sizeof(long) ist 2er Potenz! */
   num = ((num + sizeof(long)-1 + 1) & ~(sizeof(long)-1)) / sizeof(long);

   memcpy(fx_toh, from, num * sizeof(long));
   fx_toh += num;
   return new_location;
}

/*------------------------------------------------------------------------------
 * Kopiert 'num' FLOATs in den neuen Heap.
 *----------------------------------------------------------------------------*/
double *fl_swap (from, num)
double *from;
long num;
{
   double *new_location = fl_toh;

   /* ACHTUNG: wenn mehrere Referenzen auf eine Fliekommazahl existieren, */
   /* dann wird die Zahl mehrmals kopiert ! --> evtl. berlauf whrend */
   /* der GC */
   /*--------*/
   if(fl_toh + num >= fl_eoh)
      Labort("Float-Heap Overflow.");

   memcpy(fl_toh, from, num * sizeof(double));
   fl_toh += num;
   return new_location;
}

/*------------------------------------------------------------------------------
 * Rettet eine LISP-Form, falls noetig, in den neuen Heap.
 *----------------------------------------------------------------------------*/
void save_form (form)
CL_FORM *form;
{
 begin:
   switch (TYPE_OF (form))
   {
   case CL_UNBOUND:
   case CL_FIXNUM:
   case CL_CHAR:
   case CL_NIL:
   case CL_DOWNFUN:
   case CL_GLOBFUN:
   case CL_CODE:
   case CL_CFILE:
   case CL_UNIQUE_TAG:
      break;
      
   /*------------------------------------------------------------------------   
    * Garbage-Collection eines LISP-Datums,                                  
    * welches durch einen Zeiger referenziert wird.                          
    * Annahme: Konstanten des Typs 'double' und 'CL_FORM', die schon zur     
    * Uebersetzungszeit in 'floats' bzw 'const_forms' angelegt wurden,       
    * haben Adressen, die kleiner sind als die Adresse von Daten, die        
    * erst zur Laufzeit angelegt wurden                                      
    *----------------------------------------------------------------------*/
   case CL_FLOAT:
   {
      double *lptr = GET_FLOAT_PTR(form);

      if(lptr < fl_heap1)       /* Konstantes Datum */
         return;
      if (lptr >= fl_heap1 + 2 * fl_heapsize)
         Labort ("Unexpected pointer out of heap.");
      if (!(old_fl_heap <= lptr && lptr < old_fl_heap + fl_heapsize))
         Labort ("Unexpected pointer in wrong heap.");

      /* ACHTUNG: wenn mehrere Referenzen auf eine Fliekommazahl existieren, */
      /* dann wird die Zahl mehrmals kopiert! --> evtl. berlauf whrend */
      /* der GC */
      
      *fl_toh = *lptr;
      GET_FLOAT_PTR(form) = fl_toh++;
      if(fl_toh >= fl_eoh)
         Labort("Float-Heap Overflow.");
      break;
   }
   default:
   {
      CL_FORM *lptr = GET_FORM (form);
      
      /*------------------------------------------------------------------------
       * Annahme:
       * - Der initialisierte Konstantenbereich wird bereits zur
       *   bersetzungszeit alloziert.
       * - Nicht initialisierte Arrays werden erst zur Laufzeit alloziert.
       * - Zur Laufzeit allozierte Speicherbereiche liegen adressenmaessig
       *   oberhalb des Programms.
       * 
       * => Zur Ueberpruefung, ob der Zeiger nicht in den Heap zeigt,
       * reicht die Abfrage:
       *----------------------------------------------------------------------*/
      if (lptr < fo_heap)       /* Konstantes Datum */
         return;
      if (lptr >= fo_heap + 2 * form_heapsize)
         Labort ("Unexpected pointer out of heap.");
      if (!(old_form_heap <= lptr && lptr < old_form_heap + form_heapsize))
         Labort ("Unexpected pointer in wrong heap.");

      /* Wenn Vorwrtsreferenz, dann Datum bereits kopiert.
       * Neue Adresse eintragen und fertig. */
      if (TYPE_OF(lptr) == GC_FORWARD)
      {
         GET_FORM(form) = GET_FORM(lptr);
         return;
      }
      
      switch (TYPE_OF (form)) {

      /* Garbage-Collection von Symbolen */
      /*---------------------------------*/
      case CL_SYMBOL:
         GET_FORM(form) = form_swap(lptr, SYM_SIZE);
         LOAD_FORWARD(GET_FORM(form), lptr);

         lptr = GET_FORM(form); /* Zeiger auf das neue Symbol */
         
         /* SYMBOL Printnamen retten, falls String im Character-Heap liegt.
          * MAKE-SYMBOL kopiert bei einem SIMPLE-STRING nicht den
          * eigentlichen String! Deshalb kann ein SYMBOL im Heap als
          * Printnamen auch einen C-String haben. */
         if (CHAR_AR(lptr) >= (char *)fx_heap1)
            CHAR_AR(lptr) = ch_swap(CHAR_AR(lptr), AR_SIZE(lptr));
         
         save_form(lptr + OFF_SYM_PLIST);
         save_form(lptr + OFF_SYM_VALUE);
         save_form(lptr + OFF_SYM_PACKAGE);
         break;

      /* Garbage-Collection von Listen */
      /*------------------------------ */
      case CL_CONS:
         COPY(lptr, form_toh);
         COPY(lptr + 1, form_toh + 1);
         
         /* Zeiger auf kopierten CONS-Knoten setzen */
         GET_FORM(form) = form_toh;
         LOAD_FORWARD(form_toh, lptr);
         form_toh += 2;

         /* Depth-First Strategie: Zuerst den CAR weiterverfolgen */
         save_form (GET_FORM (form));

         /* Dann den CDR weiterverfolgen.
          * Hier wird 'form' neu gesetzt und durch einen Sprung an den
          * Anfang der Funktion unnoetige Rekursion vermieden. */
         form = GET_FORM (form) + 1;
         goto begin;
         break;

      /* Bei einem VECTOR wird
       * 1. Der Header kopiert, der Zeiger auf den neuen Header gesetzt und
       *    im alten Header eine Vorwrtsreferenz auf den neuen Header
       *    eingetragen. (Der Header beginnt 2 Eintrge vor 'lptr' und
       *    ist 4 Eintraege lang.)
       * 2. Falls es sich um einen DISPLACED-TO Vektor handelt, wird der
       *    Vektor, auf welchen referenziert wird, gerettet.
       * 3. Sonst wird die Lnge des Vektors bestimmt,
       *    (FILL-POINTER wird ignoriert)
       *    und die Elemente werden kopiert. */
      /*-------------------------------------*/
      case CL_VEC_T:
      case CL_VEC_FIXNUM:
      case CL_VEC_FLOAT:
      case CL_STRING:
         GET_FORM(form) = form_swap(lptr - 2, 4L) + 2;
         LOAD_FORWARD(GET_FORM(form), lptr);

         lptr = GET_FORM (form); /* Zeiger auf den neuen Header */
         if (DISPLACED_P (lptr))
         {
            save_form (DISPLACED_TO (lptr));
         }
         else
         {
            long i, size;

            size = HAS_FILL_PTR (lptr) ? AR_SIZE_WHEN_FP(lptr) : AR_SIZE(lptr);
            switch (TYPE_OF (form)) {
            case CL_VEC_T:
               FORM_AR(lptr) = form_swap(FORM_AR(lptr), size);
               for (lptr = FORM_AR(lptr), i = 0; i < size; i++)
                  save_form(lptr + i);
               break;
            case CL_VEC_FIXNUM:
               FIXNUM_AR(lptr) = fx_swap(FIXNUM_AR(lptr), size);
               break;
            case CL_VEC_FLOAT:
               FLOAT_AR(lptr) = fl_swap(FLOAT_AR(lptr), size);
               break;
            case CL_STRING:
               CHAR_AR(lptr) = ch_swap(CHAR_AR(lptr), size);
            }
         }
         break;

      /* Bei einem SIMPLE-VECTOR wird:
       * 1. Der Header kopiert.
       * 2. Der eigentliche Vektor kopiert.
       * 3. Im alten Header eine Vorwaertsreferenz auf den neuen Header
       *     eingetragen.
       * 4. Falls die Elemente des Vektors wieder CL_FORM's sind,
       *    das GC auf die Elemente ausgedehnt. */
      /*----------------------------------------*/
      case CL_SMVEC_T:
      case CL_CLOSURE:
      case CL_STRUCT:
      case CL_INSTANCE:
      {
         long i, size = AR_SIZE (lptr);

         GET_FORM(form) = form_swap(lptr, 1 + size);
         LOAD_FORWARD(GET_FORM(form), lptr);
         for (lptr = GET_FORM(lptr) + 1, i = 0; i < size; i++)
            save_form (lptr + i);
         break;
      }
      case CL_SMVEC_FIXNUM:
         GET_FORM (form) = form_swap (lptr, 2L);
         LOAD_FORWARD (GET_FORM (form), lptr);
         lptr = GET_FORM (form); /* Zeiger auf den neuen Header */
         FIXNUM_AR (lptr) = fx_swap (FIXNUM_AR (lptr), AR_SIZE (lptr));
         break;
      case CL_SMVEC_FLOAT: 
         GET_FORM (form) = form_swap (lptr, 2L);
         LOAD_FORWARD (GET_FORM (form), lptr);
         lptr = GET_FORM (form); /* Zeiger auf den neuen Header */
         FLOAT_AR (lptr) = fl_swap (FLOAT_AR (lptr), AR_SIZE (lptr));
         break;
      case CL_SMSTR:    
         GET_FORM (form) = form_swap (lptr, 2L);
         LOAD_FORWARD (GET_FORM (form), lptr);
         lptr = GET_FORM (form); /* Zeiger auf den neuen Header */
         CHAR_AR (lptr)  = ch_swap (CHAR_AR (lptr), AR_SIZE (lptr));
         break;

      /* Simple Arrays */
      /*---------------*/
      case CL_SMAR_T:  
      {
         long i, size = AR_SIZE(lptr);

         GET_FORM(form) = form_swap(lptr - 1, 2L + size) + 1;
         LOAD_FORWARD(GET_FORM (form), lptr);

         lptr = GET_FORM (form); /* Zeiger auf den neuen Header */
         AR_DIMS (lptr) = fx_swap (AR_DIMS (lptr), AR_RANK (lptr));

         lptr++;                /* Zeiger auf das eigentliche Array */
         for (i = 0; i < size; i++)
            save_form (lptr + i);
         break;
      }
      
      case CL_SMAR_FIXNUM:
      case CL_SMAR_FLOAT:
      case CL_SMAR_CHAR:
         GET_FORM (form) = form_swap (lptr - 1, 3L) + 1;
         LOAD_FORWARD (GET_FORM (form), lptr);

         lptr = GET_FORM (form); /* Zeiger auf den neuen Header */
         AR_DIMS (lptr) = fx_swap (AR_DIMS (lptr), AR_RANK (lptr));
         switch (TYPE_OF (form)) {
         case CL_SMAR_FIXNUM:
            FIXNUM_AR (lptr) = fx_swap (FIXNUM_AR (lptr), AR_SIZE (lptr));
            break;
         case CL_SMAR_FLOAT: 
            FLOAT_AR (lptr) = fl_swap (FLOAT_AR (lptr), AR_SIZE (lptr));
            break;
         case CL_SMAR_CHAR:  
            CHAR_AR (lptr)  = ch_swap (CHAR_AR (lptr), AR_SIZE (lptr));
         }
         break;

      /* Arrays */
      /*--------*/
      case CL_AR_T:
      case CL_AR_FIXNUM:
      case CL_AR_FLOAT:
      case CL_AR_CHAR:
         GET_FORM (form) = form_swap (lptr - 2, 4L) + 2;
         LOAD_FORWARD (GET_FORM (form), lptr);

         lptr = GET_FORM (form); /* Zeiger auf den neuen Header */
         AR_DIMS (lptr) = fx_swap (AR_DIMS (lptr), AR_RANK (lptr));
         if (DISPLACED_P (lptr))
         {
            save_form (DISPLACED_TO (lptr));
         }
         else
         {
            switch (TYPE_OF (form))
            {
            case CL_AR_T:
            {
               long i, size = AR_SIZE (lptr);
               
               FORM_AR (lptr) = form_swap (FORM_AR (lptr), size);
               
               /* Zeiger auf das eigentliche Array */
               for (lptr = FORM_AR (lptr), i = 0; i < size; i++)
                  save_form (lptr + i);
               break;
            }
            case CL_AR_FIXNUM:
               FIXNUM_AR (lptr) = fx_swap (FIXNUM_AR (lptr), AR_SIZE (lptr));
               break;
            case CL_AR_FLOAT:
               FLOAT_AR (lptr) = fl_swap (FLOAT_AR (lptr), AR_SIZE (lptr));
               break;
            case CL_AR_CHAR:
               CHAR_AR (lptr) = ch_swap (CHAR_AR (lptr), AR_SIZE (lptr));
            }
         }
         break;
         
      /*--------------------------------------------------*/
      case CL_IND:
         GET_FORM (form) = form_swap (lptr, 1L);
         LOAD_FORWARD (GET_FORM (form), lptr);
         form = GET_FORM (form);
         goto begin;

      /* Foreign-Datentypen */
      /*--------------------*/
      case CL_C_CHAR:
      case CL_C_UNSIGNED_CHAR:
      case CL_C_SHORT:
      case CL_C_INT:
      case CL_C_LONG:
      case CL_C_UNSIGNED_SHORT:
      case CL_C_UNSIGNED_INT:
      case CL_C_UNSIGNED_LONG:
         break;

      default:
         fprintf (stderr, ";;; Unkonwn data type %d by GC\n", TYPE_OF (form));
         exit (1);
      }
   }
   } /*switch*/
}

/*------------------------------------------------------------------------------
 * Garbage-Collector
 *----------------------------------------------------------------------------*/
void gc (base)
CL_FORM *base;
{
   CL_FORM *form_tmp = form_heap;
   long    *fx_tmp = fx_heap;
   double  *fl_tmp = fl_heap;

   CL_FORM *lptr;
   double fo_room, fx_room, fl_room;

   fprintf (stderr, ";;; GC -- ");
   fflush(stderr);

   /* Vertauschen der Heap-Bereiche */
   /* ----------------------------- */
   form_heap = old_form_heap;
   old_form_heap = form_tmp;
   form_toh = form_heap;
   form_eoh = form_heap + form_heapsize;

   fx_heap = old_fx_heap;
   old_fx_heap = fx_tmp;
   fx_toh = fx_heap;
   fx_eoh = fx_heap + fx_heapsize;

   fl_heap = old_fl_heap;
   old_fl_heap = fl_tmp;
   fl_toh = fl_heap;
   fl_eoh = fl_heap + fl_heapsize;

   /* Traversieren der Wurzelbereiche im Hauptmodul und den importierten
    * Modulen
    *--------------------------------------------------------------------*/
   gc_symbols(Ssys);
   gc_main();

   /* Traversieren des LISP-Laufzeitstacks */
   /* ------------------------------------ */
   for (lptr = stack; lptr < base; lptr++)
      save_form (lptr);
   
   /* Traversieren des Binding-Stacks */
   /* ------------------------------- */
   /* 'bind_top' zeigt auf den naechsten freien Eintrag. */
   for (lptr = bind_stack; lptr < bind_top; lptr++)
      save_form (lptr);

   fo_room = (double)(form_toh - form_heap) / form_heapsize;
   fx_room = (double)(fx_toh - fx_heap) / fx_heapsize;
   fl_room = (double)(fl_toh - fl_heap) / fl_heapsize;
   fprintf(stderr, "Allocated: %5.2f%% Forms, %5.2f%% \
Characters/Fixnums, %5.2f%% Floats \n",
           100 * fo_room, 100 * fx_room, 100 * fl_room);
}

/*------------------------------------------------------------------------------
 * Traversiert die Symbole eines Moduls
 *----------------------------------------------------------------------------*/
void gc_symbols (symbol)
CL_INIT *symbol;
{
   while (TYPE_OF (symbol))
   {
      save_form ((CL_FORM *)(symbol + OFF_SYM_PLIST));
      save_form ((CL_FORM *)(symbol + OFF_SYM_VALUE));
      save_form ((CL_FORM *)(symbol + OFF_SYM_PACKAGE));
      symbol += SYM_SIZE;
   }
}

/*------------------------------------------------------------------------------
 * Alloziert Speicherplatz fuer 'num' Integers im Heap.
 * Annahme: Integer benoetigen gerade Anzahl von Bytes.
 * Rueckgabewert: Zeiger -> Beginn des allozierten Speicherbereichs
 *----------------------------------------------------------------------------*/
long *fixnum_alloc (base, num)
CL_FORM *base;
long num;
{
   long *ptr;

   /* Falls der Heap voll ist, Garbage-Collector starten */
   /* -------------------------------------------------- */
   if (fx_toh + num > fx_eoh)
   {
      gc (base);
      if (fx_toh + num > fx_eoh)
      {
         Labort("Not enough memory for heap allocation of FIXNUM");
      }
   }

   ptr = fx_toh;
   fx_toh += num;   /* Neuen TOH setzen */
   return ptr;
}

/*------------------------------------------------------------------------------
 * Alloziert Speicherplatz fuer 'num' + 1 Characters im Heap.
 * Jeder Lisp-String wird mit einem 0 Character abgeschlossen, damit eine
 * problemlose Uebertragung nach C moeglich ist.
 * Rueckgabewert: Zeiger auf Beginn des allozierten Speicherbereichs
 *----------------------------------------------------------------------------*/
char *char_alloc (base, rnum)
CL_FORM *base;
unsigned long rnum;
{
   char *ptr; 
   long num;

   /* Anzahl der LONG's (+ 1 fuer abschliessendes \0)*/
   num = (rnum + sizeof(long)) / sizeof(long);

   /* Falls der Heap voll ist, Garbage-Collector starten */
   /*----------------------------------------------------*/
   if (fx_toh + num > fx_eoh)
   {
      gc (base);
      if (fx_toh + num > fx_eoh)
      {
         Labort("Not enough memory for heap allocation of CHARACTER");
      }
   }

   ptr = (char *)fx_toh;
   ptr[rnum] = '\0';            /* Abschliessende 0 setzen */
   fx_toh += num;               /* Neuen Top of Heap setzen */
   return ptr;
}

/*------------------------------------------------------------------------------
 * Alloziert Speicherplatz fuer <num> Floats im Heap.
 * Rueckgabewert: Zeiger -> Beginn des allozierten Speicherbereichs
 *----------------------------------------------------------------------------*/
double *float_alloc (base, num)
CL_FORM *base;
long num;
{
   double *ptr;

   /* Falls der Heap voll ist, Garbage-Collector starten */
   /* -------------------------------------------------- */
   if (fl_toh + num > fl_eoh)
   {
      gc (base);
      if (fl_toh + num > fl_eoh)
      {
         Labort("Not enough memory for heap allocation of FLOAT");
      }
   }

   ptr = fl_toh;
   fl_toh += num;           /* Top of Heap erhoehen */
   return ptr;             /* Zeiger auf erste allozierte FLOAT zurueckgeben */
}

/*------------------------------------------------------------------------------
 * Alloziert Speicherplatz fuer <num> CL_FORM's im Heap.
 * Rueckgabewert: Zeiger -> Beginn des allozierten Speicherbereichs
 *----------------------------------------------------------------------------*/
CL_FORM *form_alloc (base, num)
CL_FORM *base;
long num;
{
   CL_FORM *lptr;

    /* Kein Platz mehr fuer <num> Forms, Garbage Collector starten */
    /* ----------------------------------------------------------- */
   if (form_toh + num > form_eoh)
   {
      gc (base);
      if(form_toh + num > form_eoh)
      {
         Labort("Not enough memory for heap allocation of FORM");
      }
   }

   lptr = form_toh;
   form_toh += num;                   /* Top of Heap (Form) neu setzen */
   return lptr;                       /* Zeiger auf erste Form zurueckgeben */
}

/*------------------------------------------------------------------------------
 * Alloziert eine Liste der Laenge <len> im Heap.
 * Rueckgabewert: Zeiger -> Beginn der Liste
 *----------------------------------------------------------------------------*/
CL_FORM *list_alloc (base, len)
CL_FORM *base;
int len;
{
   CL_FORM *list, *lptr;
   int i;

   /* Listenvektor allozieren (1 CONS = 2 CL_FORM's) */
   /*------------------------------------------------*/
   list = form_alloc(base, 2L * len);

   /* Liste verketten */
   /* --------------- */
   for (lptr = list + 1, i = 0; i < len - 1; i++, lptr += 2)
      LOAD_CONS (lptr + 1, lptr);
   LOAD_NIL (lptr);

   /* Zeiger auf erste Form der Liste zurueckgeben */
   /* -------------------------------------------- */
   return list;
}

/*------------------------------------------------------------------------------
 * Gibt eine Fehlermeldung aus und bricht die Programmausfuehrung ab.
 *----------------------------------------------------------------------------*/
void Labort (msg)
char *msg;
{
   fprintf(stderr, "\n*** Abort: %s\n", msg);
   exit(1);
}

/*------------------------------------------------------------------------------
 * Gibt eine Laufzeitfehlermeldung aus und bricht ab.
 * Parameter: base ist einer Zeiger in den Lisp-Stack, an die Position des
 *                 fehlerhaften Ausdrucks
 *            msg ist ein Format-String fuer die Lisp-Funktion format
 *                als C-string
 *----------------------------------------------------------------------------*/
void Lerror (base, msg)
CL_FORM *base;
char *msg;
{
   make_string(STACK(base, 1), msg);
   COPY(STACK(base, 0), STACK(base, 2));
   Ferror(STACK(base, 1), 2);
}

/*------------------------------------------------------------------------------
 * Gibt eine Warnung zur Laufzeit aus.
 *----------------------------------------------------------------------------*/
void warning (warning_str)
char *warning_str;
{
   fprintf(stderr, ";;; Warning: %s\n", warning_str);
}
